// Copyright 2003 Regents of the University of California

// SETI_BOINC is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2, or (at your option) any later
// version.

// SETI_BOINC is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.

// You should have received a copy of the GNU General Public License along
// with SETI_BOINC; see the file COPYING.  If not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "diagnostics.h"
#include "util.h"
#include "s_util.h"
#ifdef BOINC_APP_GRAPHICS
#include "graphics_api.h"
#ifdef DYNAMIC_GRAPHICS
#include "graphics_lib.h"
#endif
#endif


#include "seti_header.h"
#include "seti.h"
#include "util.h"
#include "filesys.h"
#include "s_util.h"
#include "analyze.h"
#include "worker.h"
#include "chirpfft.h"

char * cfft_file = NULL;

int GenChirpFftPairs(
  ChirpFftPair_t ** ChirpFftPairs,
  double * MinChirpStep
) {

  long i, j, NumChirpFftPairs=0, Direction=1;
  double CRate=0.0, ChirpEquality, NextCRateThisFftLength;
  long  NumLimits = (long)swi.analysis_cfg.chirps.size();
  double * ChirpSteps ;
  double * LastRateThisFft;
  ChirpFftPair_t * ChirpFftPair = NULL;

  // Two arrays associated with the FFT length array.
  ChirpSteps      = (double *)calloc(swi.num_fft_lengths, sizeof(double));
  LastRateThisFft = (double *)calloc(swi.num_fft_lengths, sizeof(double));

  // Each FFT length has a (different) associated chirp step.
  CalcChirpSteps(ChirpSteps, MinChirpStep);
  
  // How close 2 chirp rates have to be to be considered equal.
  // We don't want small floating point dicrepensies to void
  // an equality.
  ChirpEquality = *MinChirpStep/100.0;

  for(i = 0; i < NumLimits || Direction == 1; i++) {

	// If done with positive chirps, do negative chirps.
	if(i == NumLimits) {
		i = 0;
	 	Direction = -1;
		// Reinit LastRateThisFft to all zeros.
		free(LastRateThisFft);
		LastRateThisFft = (double *)calloc(swi.num_fft_lengths, sizeof(double));
		// start the negatives
		CRate = 0.0 - *MinChirpStep;
	}

  	while(CRate * Direction <= swi.analysis_cfg.chirps[i].chirp_limit) {

		for(j = 0; j < swi.num_fft_lengths; j++) {

			NextCRateThisFftLength = LastRateThisFft[j] + ChirpSteps[j] * Direction;

			// If we want to do this chirp rate for this FFT length  
			if((swi.analysis_fft_lengths[j] 		& 
			    swi.analysis_cfg.chirps[i].fft_len_flags)	&& 
  			// and we are at chirp 0 or have advanced a chirp step 
			// for this FFT length...
			   (CRate == 0 					|| 
			    VeryClose(CRate, NextCRateThisFftLength, ChirpEquality))) {

				// ...then add this pair,
			 	ChirpFftPair = (ChirpFftPair_t *)realloc(
                                	           (void *)ChirpFftPair, 
   			                	   sizeof(ChirpFftPair_t)*(NumChirpFftPairs+1));
    			 	if (ChirpFftPair == NULL) SETIERROR(MALLOC_FAILED, "ChirpFftPair == NULL");
			 	ChirpFftPair[NumChirpFftPairs].ChirpRate = (float)CRate;
    			 	ChirpFftPair[NumChirpFftPairs].FftLen = swi.analysis_fft_lengths[j];
				// count it,
    				NumChirpFftPairs++;
				// and bump up the chirp rate for this FFT length
				LastRateThisFft[j] = CRate;
			}

		} // end for loop through FFT lengths

		CRate += *MinChirpStep * Direction;

	} // end while loop through chirp rates

  } // end for loop through cfft parameter table

  *ChirpFftPairs = ChirpFftPair;
  if (ChirpSteps) free (ChirpSteps);
  if (LastRateThisFft) free (LastRateThisFft);
  return(NumChirpFftPairs);
}


void CalcChirpSteps(double*  ChirpSteps, double* MinChirpStep) {

	double ChirpRes   = swi.analysis_cfg.chirp_resolution;
	double NumSamples = swi.nsamples;
	double SampleTime = (double)(swi.nsamples/swi.subband_sample_rate);
	

	int i;
  	double SecondsInFft;
	long FftLen;

	// For each FFT length, calculate the corresponding chirp 
	// step (in Hz/s), given the chirp resolution (in FFT bins).
	// Keep track of the minimum chirp step.
	*MinChirpStep = 0.0;
  	for(i = 0; i < swi.num_fft_lengths; i++) {
		FftLen = swi.analysis_fft_lengths[i];
    		SecondsInFft =  SampleTime / (NumSamples / FftLen);
    		ChirpSteps[i] = ChirpRes / (SecondsInFft * SecondsInFft);
		if(ChirpSteps[i] < *MinChirpStep || *MinChirpStep == 0.0) {
			*MinChirpStep = ChirpSteps[i];
		}
  	}
}


bool VeryClose(double a, double b, double CloseEnough) {

  double Diff;

  Diff = fabs(a - b);

  if(Diff > CloseEnough)
    return(false);
  else
    return(true);
}


int ReadCFftFile(ChirpFftPair_t ** ChirpFftPairs, double * MinChirpStep) {

  //  ChirpFftSet_t ChirpFftSet;
  ChirpFftPair_t * ChirpFftPair;
  int NumChirpFftPairs;
  FILE * cfftfp;
  float ChirpRate;
  int FftLen;

  NumChirpFftPairs = 0;

  cfftfp = boinc_fopen(cfft_file, "r");

  ChirpFftPair = (ChirpFftPair_t*) calloc(1,sizeof(ChirpFftPair_t));

  while(fscanf(cfftfp, "%f %d", &ChirpRate, &FftLen) == 2) {
    ChirpFftPair = (ChirpFftPair_t *)realloc(
                     (void *)ChirpFftPair, sizeof(ChirpFftPair_t)*(NumChirpFftPairs+1));
    if (ChirpFftPair == NULL) SETIERROR(MALLOC_FAILED, "ChirpFftPair == NULL");
    ChirpFftPair[NumChirpFftPairs].ChirpRate = ChirpRate;
    ChirpFftPair[NumChirpFftPairs].FftLen    = FftLen;
    NumChirpFftPairs++;
  }

  *ChirpFftPairs = ChirpFftPair;
  *MinChirpStep  = (float) TEST_MIN_CHIRP_STEP;

  fclose(cfftfp);

  return(NumChirpFftPairs);
}

